@@ -3,29 +3,31 @@
 import datetime
 from itertools import chain
 from decimal import Decimal
+from sql import Table, Literal
+from sql.functions import Overlay, Position
+from sql.aggregate import Count
+
 from trytond.model import Workflow, ModelView, ModelSQL, fields
 from trytond.modules.company import CompanyReport
 from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
     Button
-from trytond.backend import TableHandler
+from trytond import backend
 from trytond.pyson import Eval, Bool, If, PYSONEncoder, Id
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 
 __all__ = ['Purchase', 'PurchaseInvoice', 'PurchaseIgnoredInvoice',
     'PurchaseRecreadtedInvoice', 'PurchaseLine', 'PurchaseLineTax',
-    'PurchaseLineIgnoredMove',
-    'PurchaseLineRecreatedMove', 'PurchaseReport', 'Template', 'Product',
-    'ProductSupplier', 'ProductSupplierPrice', 'ShipmentIn',
-    'ShipmentInReturn', 'Move', 'OpenSupplier', 'HandleShipmentExceptionAsk',
-    'HandleShipmentException', 'HandleInvoiceExceptionAsk',
-    'HandleInvoiceException']
+    'PurchaseLineIgnoredMove', 'PurchaseLineRecreatedMove', 'PurchaseReport',
+    'OpenSupplier', 'HandleShipmentExceptionAsk', 'HandleShipmentException',
+    'HandleInvoiceExceptionAsk', 'HandleInvoiceException']
 __metaclass__ = PoolMeta
 
 _STATES = {
     'readonly': Eval('state') != 'draft',
     }
 _DEPENDS = ['state']
+_ZERO = Decimal(0)
 
 
 class Purchase(Workflow, ModelSQL, ModelView):
@@ -38,7 +40,7 @@ class Purchase(Workflow, ModelSQL, ModelView):
             },
         domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', 0)),
+                Eval('context', {}).get('company', -1)),
             ],
         depends=['state'], select=True)
     reference = fields.Char('Reference', size=None, readonly=True, select=True)
@@ -85,19 +87,19 @@ class Purchase(Workflow, ModelSQL, ModelView):
     comment = fields.Text('Comment')
     untaxed_amount = fields.Function(fields.Numeric('Untaxed',
             digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_untaxed_amount')
+            depends=['currency_digits']), 'get_amount')
     untaxed_amount_cache = fields.Numeric('Untaxed Cache',
         digits=(16, Eval('currency_digits', 2)),
         depends=['currency_digits'])
     tax_amount = fields.Function(fields.Numeric('Tax',
             digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_tax_amount')
+            depends=['currency_digits']), 'get_amount')
     tax_amount_cache = fields.Numeric('Tax Cache',
         digits=(16, Eval('currency_digits', 2)),
         depends=['currency_digits'])
     total_amount = fields.Function(fields.Numeric('Total',
             digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_total_amount')
+            depends=['currency_digits']), 'get_amount')
     total_amount_cache = fields.Numeric('Total Cache',
         digits=(16, Eval('currency_digits', 2)),
         depends=['currency_digits'])
@@ -193,18 +195,34 @@ class Purchase(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        model_data = Table('ir_model_data')
+        model_field = Table('ir_model_field')
+        sql_table = cls.__table__()
+
         # Migration from 1.2: packing renamed into shipment
-        cursor.execute("UPDATE ir_model_data "
-            "SET fs_id = REPLACE(fs_id, 'packing', 'shipment') "
-            "WHERE fs_id like '%%packing%%' AND module = %s",
-            (module_name,))
-        cursor.execute("UPDATE ir_model_field "
-            "SET relation = REPLACE(relation, 'packing', 'shipment'), "
-                "name = REPLACE(name, 'packing', 'shipment') "
-            "WHERE (relation like '%%packing%%' "
-                "OR name like '%%packing%%') AND module = %s",
-            (module_name,))
+        cursor.execute(*model_data.update(
+                columns=[model_data.fs_id],
+                values=[Overlay(model_data.fs_id, 'shipment',
+                        Position('packing', model_data.fs_id),
+                        len('packing'))],
+                where=model_data.fs_id.like('%packing%')
+                & (model_data.module == module_name)))
+        cursor.execute(*model_field.update(
+                columns=[model_field.relation],
+                values=[Overlay(model_field.relation, 'shipment',
+                        Position('packing', model_field.relation),
+                        len('packing'))],
+                where=model_field.relation.like('%packing%')
+                & (model_field.module == module_name)))
+        cursor.execute(*model_field.update(
+                columns=[model_field.name],
+                values=[Overlay(model_field.name, 'shipment',
+                        Position('packing', model_field.name),
+                        len('packing'))],
+                where=model_field.name.like('%packing%')
+                & (model_field.module == module_name)))
         table = TableHandler(cursor, cls, module_name)
         table.column_rename('packing_state', 'shipment_state')
 
@@ -212,9 +230,10 @@ class Purchase(Workflow, ModelSQL, ModelView):
 
         # Migration from 1.2: rename packing to shipment in
         # invoice_method values
-        cursor.execute("UPDATE " + cls._table + " "
-            "SET invoice_method = 'shipment' "
-            "WHERE invoice_method = 'packing'")
+        cursor.execute(*sql_table.update(
+                columns=[sql_table.invoice_method],
+                values=['shipment'],
+                where=sql_table.invoice_method == 'packing'))
 
         table = TableHandler(cursor, cls, module_name)
         # Migration from 2.2: warehouse is no more required
@@ -285,6 +304,7 @@ class Purchase(Workflow, ModelSQL, ModelView):
         PaymentTerm = pool.get('account.invoice.payment_term')
         Currency = pool.get('currency.currency')
         cursor = Transaction().cursor
+        table = self.__table__()
         changes = {
             'invoice_address': None,
             'payment_term': None,
@@ -298,13 +318,13 @@ class Purchase(Workflow, ModelSQL, ModelView):
             if self.party.supplier_payment_term:
                 payment_term = self.party.supplier_payment_term
 
-            subquery = cursor.limit_clause('SELECT currency '
-                'FROM "' + self._table + '" '
-                'WHERE party = %s '
-                'ORDER BY id DESC', 10)
-            cursor.execute('SELECT currency FROM (' + subquery + ') AS p '
-                'GROUP BY currency '
-                'ORDER BY COUNT(1) DESC', (self.party.id,))
+            subquery = table.select(table.currency,
+                where=table.party == self.party.id,
+                order_by=table.id,
+                limit=10)
+            cursor.execute(*subquery.select(subquery.currency,
+                    group_by=subquery.currency,
+                    order_by=Count(Literal(1)).desc))
             row = cursor.fetchone()
             if row:
                 currency_id, = row
@@ -387,25 +407,11 @@ class Purchase(Workflow, ModelSQL, ModelView):
                 changes['total_amount'])
         return changes
 
-    def get_untaxed_amount(self, name):
-        '''
-        Return the untaxed amount for each purchases
-        '''
-        if (self.state in self._states_cached
-                and self.untaxed_amount_cache is not None):
-            return self.untaxed_amount_cache
-        amount = sum((l.amount for l in self.lines
-                if l.type == 'line'), Decimal(0))
-        return self.currency.round(amount)
-
-    def get_tax_amount(self, name):
+    def get_tax_amount(self):
         pool = Pool()
         Tax = pool.get('account.tax')
         Invoice = pool.get('account.invoice')
 
-        if (self.state in self._states_cached
-                and self.tax_amount_cache is not None):
-            return self.tax_amount_cache
         context = self.get_tax_context()
         taxes = {}
         for line in self.lines:
@@ -421,15 +427,39 @@ class Purchase(Workflow, ModelSQL, ModelView):
                     taxes[key] = val['amount']
                 else:
                     taxes[key] += val['amount']
-        amount = sum((self.currency.round(tax) for tax in taxes.itervalues()),
-            Decimal(0))
-        return self.currency.round(amount)
+        return sum((self.currency.round(tax) for tax in taxes.values()), _ZERO)
 
-    def get_total_amount(self, name):
-        if (self.state in self._states_cached
-                and self.total_amount_cache is not None):
-            return self.total_amount_cache
-        return self.currency.round(self.untaxed_amount + self.tax_amount)
+    @classmethod
+    def get_amount(cls, purchases, names):
+        untaxed_amount = {}
+        tax_amount = {}
+        total_amount = {}
+
+        for purchase in purchases:
+            if (purchase.state in cls._states_cached
+                    and purchase.untaxed_amount_cache is not None
+                    and purchase.tax_amount_cache is not None
+                    and purchase.total_amount_cache is not None):
+                untaxed_amount[purchase.id] = purchase.untaxed_amount_cache
+                tax_amount[purchase.id] = purchase.tax_amount_cache
+                total_amount[purchase.id] = purchase.total_amount_cache
+            else:
+                untaxed_amount[purchase.id] = sum(
+                    (line.amount for line in purchase.lines
+                        if line.type == 'line'), _ZERO)
+                tax_amount[purchase.id] = purchase.get_tax_amount()
+                total_amount[purchase.id] = (
+                    untaxed_amount[purchase.id] + tax_amount[purchase.id])
+
+        result = {
+            'untaxed_amount': untaxed_amount,
+            'tax_amount': tax_amount,
+            'total_amount': total_amount,
+            }
+        for key in result.keys():
+            if key not in names:
+                del result[key]
+        return result
 
     def get_invoice_state(self):
         '''
@@ -604,7 +634,6 @@ class Purchase(Workflow, ModelSQL, ModelView):
             return Invoice(
                 company=self.company,
                 type=invoice_type,
-                reference=self.reference,
                 journal=journal,
                 party=self.party,
                 invoice_address=self.invoice_address,
@@ -661,7 +690,6 @@ class Purchase(Workflow, ModelSQL, ModelView):
         with Transaction().set_user(0, set_context=True):
             return ShipmentInReturn(
                 company=self.company,
-                reference=self.reference,
                 from_location=self.warehouse.storage_location,
                 to_location=self.party.supplier_location,
                 )
@@ -788,9 +816,7 @@ class PurchaseLine(ModelSQL, ModelView):
     _rec_name = 'description'
     purchase = fields.Many2One('purchase.purchase', 'Purchase',
             ondelete='CASCADE', select=True, required=True)
-    sequence = fields.Integer('Sequence',
-        order_field='(%(table)s.sequence IS NULL) %(order)s, '
-        '%(table)s.sequence %(order)s')
+    sequence = fields.Integer('Sequence')
     type = fields.Selection([
         ('line', 'Line'),
         ('subtotal', 'Subtotal'),
@@ -838,7 +864,6 @@ class PurchaseLine(ModelSQL, ModelView):
                 []),
             'stock_date_end': Eval('_parent_purchase', {}).get(
                 'purchase_date'),
-            'purchasable': True,
             'stock_skip_warehouse': True,
             }, depends=['type'])
     product_uom_category = fields.Function(
@@ -900,29 +925,37 @@ class PurchaseLine(ModelSQL, ModelView):
         super(PurchaseLine, cls).__setup__()
         cls._order.insert(0, ('sequence', 'ASC'))
         cls._error_messages.update({
-                'supplier_location_required': ('Purchase "%(purchase)s" misses '
-                    'the supplier location for line "%(line)s".'),
-                'missing_account_expense': ('Product "%(product)s" of purchase '
-                    '%(purchase)s misses an expense account.'),
+                'supplier_location_required': ('Purchase "%(purchase)s" '
+                    'misses the supplier location for line "%(line)s".'),
+                'missing_account_expense': ('Product "%(product)s" of '
+                    'purchase %(purchase)s misses an expense account.'),
                 'missing_account_expense_property': ('Purchase "%(purchase)s" '
                     'misses an "account expense" default property.'),
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        super(PurchaseLine, cls).__register__(module_name)
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
+        super(PurchaseLine, cls).__register__(module_name)
         table = TableHandler(cursor, cls, module_name)
 
         # Migration from 1.0 comment change into note
         if table.column_exist('comment'):
-            cursor.execute('UPDATE "' + cls._table + '" '
-                'SET note = comment')
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.note],
+                    values=[sql_table.comment]))
             table.drop_column('comment', exception=True)
 
         # Migration from 2.4: drop required on sequence
         table.not_null_action('sequence', action='remove')
 
+    @staticmethod
+    def order_sequence(tables):
+        table, _ = tables[None]
+        return [table.sequence == None, table.sequence]
+
     @staticmethod
     def default_type():
         return 'line'
@@ -1006,7 +1039,7 @@ class PurchaseLine(ModelSQL, ModelView):
 
         with Transaction().set_context(self._get_context_purchase_price()):
             res['unit_price'] = Product.get_purchase_price([self.product],
-                self.quantity or 0)[self.product.id]
+                abs(self.quantity or 0))[self.product.id]
             if res['unit_price']:
                 res['unit_price'] = res['unit_price'].quantize(
                     Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
@@ -1046,7 +1079,7 @@ class PurchaseLine(ModelSQL, ModelView):
 
         with Transaction().set_context(self._get_context_purchase_price()):
             res['unit_price'] = Product.get_purchase_price([self.product],
-                self.quantity or 0)[self.product.id]
+                abs(self.quantity or 0))[self.product.id]
             if res['unit_price']:
                 res['unit_price'] = res['unit_price'].quantize(
                     Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
@@ -1166,6 +1199,7 @@ class PurchaseLine(ModelSQL, ModelView):
         invoice_line.product = self.product
         invoice_line.unit_price = self.unit_price
         invoice_line.taxes = self.taxes
+        invoice_line.invoice_type = invoice_type
         if self.product:
             invoice_line.account = self.product.account_expense_used
             if not invoice_line.account:
@@ -1282,531 +1316,6 @@ class PurchaseReport(CompanyReport):
     __name__ = 'purchase.purchase'
 
 
-class Template:
-    __name__ = "product.template"
-    purchasable = fields.Boolean('Purchasable', states={
-            'readonly': ~Eval('active', True),
-            }, depends=['active'])
-    product_suppliers = fields.One2Many('purchase.product_supplier',
-        'product', 'Suppliers', states={
-            'readonly': ~Eval('active', True),
-            'invisible': (~Eval('purchasable', False)
-                | ~Eval('context', {}).get('company', 0)),
-            }, depends=['active', 'purchasable'])
-    purchase_uom = fields.Many2One('product.uom', 'Purchase UOM', states={
-            'readonly': ~Eval('active'),
-            'invisible': ~Eval('purchasable'),
-            'required': Eval('purchasable', False),
-            },
-        domain=[('category', '=', Eval('default_uom_category'))],
-        on_change_with=['default_uom', 'purchase_uom', 'purchasable'],
-        depends=['active', 'purchasable', 'default_uom_category'])
-
-    @classmethod
-    def __setup__(cls):
-        super(Template, cls).__setup__()
-        cls._error_messages.update({
-                'change_purchase_uom': ('Purchase prices are based '
-                    'on the purchase uom, are you sure to change it?'),
-                })
-        required = ~Eval('account_category') & Eval('purchasable', False)
-        if not cls.account_expense.states.get('required'):
-            cls.account_expense.states['required'] = required
-        else:
-            cls.account_expense.states['required'] = (
-                cls.account_expense.states['required'] | required)
-        if 'account_category' not in cls.account_expense.depends:
-            cls.account_expense.depends.append('account_category')
-        if 'purchasable' not in cls.account_expense.depends:
-            cls.account_expense.depends.append('purchasable')
-
-    @staticmethod
-    def default_purchasable():
-        return Transaction().context.get('purchasable') or False
-
-    def on_change_with_purchase_uom(self):
-        if self.default_uom:
-            if self.purchase_uom:
-                if self.default_uom.category == self.purchase_uom.category:
-                    return self.purchase_uom.id
-                else:
-                    return self.default_uom.id
-            else:
-                return self.default_uom.id
-
-    @classmethod
-    def write(cls, templates, vals):
-        if vals.get("purchase_uom"):
-            for template in templates:
-                if not template.purchase_uom:
-                    continue
-                if template.purchase_uom.id == vals["purchase_uom"]:
-                    continue
-                for product in template.products:
-                    if not product.product_suppliers:
-                        continue
-                    cls.raise_user_warning(
-                            '%s@product_template' % template.id,
-                            'change_purchase_uom')
-        super(Template, cls).write(templates, vals)
-
-
-class Product:
-    __name__ = 'product.product'
-
-    @classmethod
-    def get_purchase_price(cls, products, quantity=0):
-        '''
-        Return purchase price for product ids.
-        The context that can have as keys:
-            uom: the unit of measure
-            supplier: the supplier party id
-            currency: the currency id for the returned price
-        '''
-        pool = Pool()
-        Uom = pool.get('product.uom')
-        User = pool.get('res.user')
-        Currency = pool.get('currency.currency')
-        Date = pool.get('ir.date')
-
-        today = Date.today()
-        res = {}
-
-        uom = None
-        if Transaction().context.get('uom'):
-            uom = Uom(Transaction().context['uom'])
-
-        currency = None
-        if Transaction().context.get('currency'):
-            currency = Currency(Transaction().context['currency'])
-
-        user = User(Transaction().user)
-
-        for product in products:
-            res[product.id] = product.cost_price
-            default_uom = product.default_uom
-            default_currency = (user.company.currency if user.company
-                else None)
-            if not uom:
-                uom = default_uom
-            if (Transaction().context.get('supplier')
-                    and product.product_suppliers):
-                supplier_id = Transaction().context['supplier']
-                for product_supplier in product.product_suppliers:
-                    if product_supplier.party.id == supplier_id:
-                        for price in product_supplier.prices:
-                            if Uom.compute_qty(product.purchase_uom,
-                                    price.quantity, uom) <= quantity:
-                                res[product.id] = price.unit_price
-                                default_uom = product.purchase_uom
-                                default_currency = product_supplier.currency
-                        break
-            res[product.id] = Uom.compute_price(default_uom, res[product.id],
-                uom)
-            if currency and default_currency:
-                date = Transaction().context.get('purchase_date') or today
-                with Transaction().set_context(date=date):
-                    res[product.id] = Currency.compute(default_currency,
-                        res[product.id], currency, round=False)
-        return res
-
-
-class ProductSupplier(ModelSQL, ModelView):
-    'Product Supplier'
-    __name__ = 'purchase.product_supplier'
-    product = fields.Many2One('product.template', 'Product', required=True,
-            ondelete='CASCADE', select=True)
-    party = fields.Many2One('party.party', 'Supplier', required=True,
-            ondelete='CASCADE', select=True, on_change=['party'])
-    name = fields.Char('Name', size=None, translate=True, select=True)
-    code = fields.Char('Code', size=None, select=True)
-    sequence = fields.Integer('Sequence',
-        order_field='(%(table)s.sequence IS NULL) %(order)s, '
-        '%(table)s.sequence %(order)s')
-    prices = fields.One2Many('purchase.product_supplier.price',
-            'product_supplier', 'Prices')
-    company = fields.Many2One('company.company', 'Company', required=True,
-        ondelete='CASCADE', select=True,
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', 0)),
-            ])
-    delivery_time = fields.Integer('Delivery Time', help="In number of days")
-    currency = fields.Many2One('currency.currency', 'Currency', required=True,
-        ondelete='RESTRICT')
-
-    @classmethod
-    def __setup__(cls):
-        super(ProductSupplier, cls).__setup__()
-        cls._order.insert(0, ('sequence', 'ASC'))
-
-    @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().cursor
-        table = TableHandler(cursor, cls, module_name)
-
-        # Migration from 2.2 new field currency
-        created_currency = table.column_exist('currency')
-
-        super(ProductSupplier, cls).__register__(module_name)
-
-        # Migration from 2.2 fill currency
-        if not created_currency:
-            Company = Pool().get('company.company')
-            limit = cursor.IN_MAX
-            cursor.execute('SELECT count(id) FROM "' + cls._table + '"')
-            product_supplier_count, = cursor.fetchone()
-            for offset in range(0, product_supplier_count, limit):
-                cursor.execute(cursor.limit_clause(
-                        'SELECT p.id, c.currency '
-                        'FROM "' + cls._table + '" AS p '
-                        'INNER JOIN "' + Company._table + '" AS c '
-                            'ON p.company = c.id '
-                        'ORDER BY p.id',
-                        limit, offset))
-                for product_supplier_id, currency_id in cursor.fetchall():
-                    cursor.execute('UPDATE "' + cls._table + '" '
-                        'SET currency = %s '
-                        'WHERE id = %s', (currency_id, product_supplier_id))
-
-        # Migration from 2.4: drop required on sequence
-        table.not_null_action('sequence', action='remove')
-
-        # Migration from 2.6: drop required on delivery_time
-        table.not_null_action('delivery_time', action='remove')
-
-    @staticmethod
-    def default_company():
-        return Transaction().context.get('company')
-
-    @staticmethod
-    def default_currency():
-        Company = Pool().get('company.company')
-        if Transaction().context.get('company'):
-            company = Company(Transaction().context['company'])
-            return company.currency.id
-
-    def on_change_party(self):
-        cursor = Transaction().cursor
-        changes = {
-            'currency': self.default_currency(),
-            }
-        if self.party:
-            cursor.execute('SELECT currency FROM "' + self._table + '" '
-                'WHERE party = %s '
-                'GROUP BY currency '
-                'ORDER BY COUNT(1) DESC', (self.party.id,))
-            row = cursor.fetchone()
-            if row:
-                changes['currency'], = row
-        return changes
-
-    def compute_supply_date(self, date=None):
-        '''
-        Compute the supply date for the Product Supplier at the given date
-        '''
-        Date = Pool().get('ir.date')
-
-        if not date:
-            date = Date.today()
-        if self.delivery_time is None:
-            return datetime.date.max
-        return date + datetime.timedelta(self.delivery_time)
-
-    def compute_purchase_date(self, date):
-        '''
-        Compute the purchase date for the Product Supplier at the given date
-        '''
-        Date = Pool().get('ir.date')
-
-        if self.delivery_time is None:
-            return Date.today()
-        return date - datetime.timedelta(self.delivery_time)
-
-
-class ProductSupplierPrice(ModelSQL, ModelView):
-    'Product Supplier Price'
-    __name__ = 'purchase.product_supplier.price'
-    product_supplier = fields.Many2One('purchase.product_supplier',
-            'Supplier', required=True, ondelete='CASCADE')
-    quantity = fields.Float('Quantity', required=True, help='Minimal quantity')
-    unit_price = fields.Numeric('Unit Price', required=True, digits=(16, 4))
-
-    @classmethod
-    def __setup__(cls):
-        super(ProductSupplierPrice, cls).__setup__()
-        cls._order.insert(0, ('quantity', 'ASC'))
-
-    @staticmethod
-    def default_quantity():
-        return 0.0
-
-
-class ShipmentIn:
-    __name__ = 'stock.shipment.in'
-
-    @classmethod
-    def __setup__(cls):
-        super(ShipmentIn, cls).__setup__()
-        add_remove = [
-            ('supplier', '=', Eval('supplier')),
-            ]
-        if not cls.incoming_moves.add_remove:
-            cls.incoming_moves.add_remove = add_remove
-        else:
-            cls.incoming_moves.add_remove = [
-                add_remove,
-                cls.incoming_moves.add_remove,
-                ]
-        if 'supplier' not in cls.incoming_moves.depends:
-            cls.incoming_moves.depends.append('supplier')
-
-        cls._error_messages.update({
-                'reset_move': ('You cannot reset to draft move "%s" because '
-                    'it was generated by a purchase.'),
-                })
-
-    @classmethod
-    def write(cls, shipments, vals):
-        pool = Pool()
-        Purchase = pool.get('purchase.purchase')
-        PurchaseLine = pool.get('purchase.line')
-
-        super(ShipmentIn, cls).write(shipments, vals)
-
-        if 'state' in vals and vals['state'] in ('received', 'cancel'):
-            purchases = []
-            move_ids = []
-            for shipment in shipments:
-                move_ids.extend([x.id for x in shipment.incoming_moves])
-
-            purchase_lines = PurchaseLine.search([
-                    ('moves', 'in', move_ids),
-                    ])
-            if purchase_lines:
-                for purchase_line in purchase_lines:
-                    if purchase_line.purchase not in purchases:
-                        purchases.append(purchase_line.purchase)
-
-            with Transaction().set_user(0, set_context=True):
-                purchases = Purchase.browse([p.id for p in purchases])
-                Purchase.process(purchases)
-
-    @classmethod
-    @ModelView.button
-    @Workflow.transition('draft')
-    def draft(cls, shipments):
-        PurchaseLine = Pool().get('purchase.line')
-        for shipment in shipments:
-            for move in shipment.incoming_moves:
-                if (move.state == 'cancel'
-                        and isinstance(move.origin, PurchaseLine)):
-                    cls.raise_user_error('reset_move', (move.rec_name,))
-
-        return super(ShipmentIn, cls).draft(shipments)
-
-
-class ShipmentInReturn:
-    __name__ = 'stock.shipment.in.return'
-
-    @classmethod
-    def __setup__(cls):
-        super(ShipmentInReturn, cls).__setup__()
-        cls._error_messages.update({
-                'reset_move': ('You cannot reset to draft a move generated '
-                    'by a purchase.'),
-                })
-
-    @classmethod
-    def write(cls, shipments, vals):
-        pool = Pool()
-        Purchase = pool.get('purchase.purchase')
-        PurchaseLine = pool.get('purchase.line')
-
-        super(ShipmentInReturn, cls).write(shipments, vals)
-
-        if 'state' in vals and vals['state'] == 'done':
-            move_ids = []
-            for shipment in shipments:
-                move_ids.extend([x.id for x in shipment.moves])
-
-            purchase_lines = PurchaseLine.search([
-                    ('moves', 'in', move_ids),
-                    ])
-            purchases = set()
-            if purchase_lines:
-                for purchase_line in purchase_lines:
-                    purchases.add(purchase_line.purchase)
-
-            with Transaction().set_user(0, set_context=True):
-                purchases = Purchase.browse([p.id for p in purchases])
-                Purchase.process(purchases)
-
-    @classmethod
-    @ModelView.button
-    @Workflow.transition('draft')
-    def draft(cls, shipments):
-        PurchaseLine = Pool().get('purchase.line')
-        for shipment in shipments:
-            for move in shipment.moves:
-                if (move.state == 'cancel'
-                        and isinstance(move.origin, PurchaseLine)):
-                    cls.raise_user_error('reset_move')
-
-        return super(ShipmentInReturn, cls).draft(shipments)
-
-
-class Move(ModelSQL, ModelView):
-    __name__ = 'stock.move'
-    purchase = fields.Function(fields.Many2One('purchase.purchase', 'Purchase',
-            states={
-                'invisible': ~Eval('purchase_visible', False),
-                },
-            depends=['purchase_visible']), 'get_purchase',
-        searcher='search_purchase')
-    purchase_quantity = fields.Function(fields.Float('Purchase Quantity',
-            digits=(16, Eval('unit_digits', 2)),
-            states={
-                'invisible': ~Eval('purchase_visible', False),
-                },
-            depends=['purchase_visible', 'unit_digits']),
-        'get_purchase_fields')
-    purchase_unit = fields.Function(fields.Many2One('product.uom',
-            'Purchase Unit', states={
-                'invisible': ~Eval('purchase_visible', False),
-                }, depends=['purchase_visible']), 'get_purchase_fields')
-    purchase_unit_digits = fields.Function(fields.Integer(
-        'Purchase Unit Digits'), 'get_purchase_fields')
-    purchase_unit_price = fields.Function(fields.Numeric('Purchase Unit Price',
-            digits=(16, 4), states={
-                'invisible': ~Eval('purchase_visible', False),
-                }, depends=['purchase_visible']), 'get_purchase_fields')
-    purchase_currency = fields.Function(fields.Many2One('currency.currency',
-            'Purchase Currency', states={
-                'invisible': ~Eval('purchase_visible', False),
-                }, depends=['purchase_visible']), 'get_purchase_fields')
-    purchase_visible = fields.Function(fields.Boolean('Purchase Visible',
-        on_change_with=['from_location']), 'on_change_with_purchase_visible')
-    supplier = fields.Function(fields.Many2One('party.party', 'Supplier'),
-        'get_supplier', searcher='search_supplier')
-    purchase_exception_state = fields.Function(fields.Selection([
-        ('', ''),
-        ('ignored', 'Ignored'),
-        ('recreated', 'Recreated'),
-        ], 'Exception State'), 'get_purchase_exception_state')
-
-    @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().cursor
-
-        super(Move, cls).__register__(module_name)
-
-        table = TableHandler(cursor, cls, module_name)
-
-        # Migration from 2.6: remove purchase_line
-        if table.column_exist('purchase_line'):
-            cursor.execute('UPDATE "' + cls._table + '" '
-                'SET origin = \'purchase.line,\' || purchase_line '
-                'WHERE purchase_line IS NOT NULL')
-            table.drop_column('purchase_line')
-
-    @classmethod
-    def _get_origin(cls):
-        models = super(Move, cls)._get_origin()
-        models.append('purchase.line')
-        return models
-
-    def get_purchase(self, name):
-        PurchaseLine = Pool().get('purchase.line')
-        if isinstance(self.origin, PurchaseLine):
-            return self.origin.purchase.id
-
-    @classmethod
-    def search_purchase(cls, name, clause):
-        return [('origin.' + name,) + tuple(clause[1:]) + ('purchase.line',)]
-
-    def get_purchase_exception_state(self, name):
-        PurchaseLine = Pool().get('purchase.line')
-        if not isinstance(self.origin, PurchaseLine):
-            return ''
-        if self in self.origin.moves_recreated:
-            return 'recreated'
-        if self in self.origin.moves_ignored:
-            return 'ignored'
-
-    def get_purchase_fields(self, name):
-        PurchaseLine = Pool().get('purchase.line')
-        if isinstance(self.origin, PurchaseLine):
-            if name[9:] == 'currency':
-                return self.origin.purchase.currency.id
-            elif name[9:] in ('quantity', 'unit_digits', 'unit_price'):
-                return getattr(self.origin, name[9:])
-            else:
-                return getattr(self.origin, name[9:]).id
-        else:
-            if name[9:] == 'quantity':
-                return 0.0
-            elif name[9:] == 'unit_digits':
-                return 2
-
-    def on_change_with_purchase_visible(self, name=None):
-        if self.from_location:
-            if self.from_location.type == 'supplier':
-                return True
-        return False
-
-    def get_supplier(self, name):
-        PurchaseLine = Pool().get('purchase.line')
-        if isinstance(self.origin, PurchaseLine):
-            return self.origin.purchase.party.id
-
-    @classmethod
-    def search_supplier(cls, name, clause):
-        return [('origin.purchase.party',) + tuple(clause[1:]) +
-            ('purchase.line',)]
-
-    @classmethod
-    def write(cls, moves, vals):
-        pool = Pool()
-        Purchase = pool.get('purchase.purchase')
-        PurchaseLine = pool.get('purchase.line')
-
-        super(Move, cls).write(moves, vals)
-        if 'state' in vals and vals['state'] in ('cancel',):
-            purchases = set()
-            purchase_lines = PurchaseLine.search([
-                    ('moves', 'in', [m.id for m in moves]),
-                    ])
-            if purchase_lines:
-                for purchase_line in purchase_lines:
-                    purchases.add(purchase_line.purchase)
-            if purchases:
-                with Transaction().set_user(0, set_context=True):
-                    purchases = Purchase.browse([p.id for p in purchases])
-                    Purchase.process(purchases)
-
-    @classmethod
-    def delete(cls, moves):
-        pool = Pool()
-        Purchase = pool.get('purchase.purchase')
-        PurchaseLine = pool.get('purchase.line')
-
-        purchases = set()
-        purchase_lines = PurchaseLine.search([
-                ('moves', 'in', [m.id for m in moves]),
-                ])
-
-        super(Move, cls).delete(moves)
-
-        if purchase_lines:
-            for purchase_line in purchase_lines:
-                purchases.add(purchase_line.purchase)
-            if purchases:
-                with Transaction().set_user(0, set_context=True):
-                    purchases = Purchase.browse([p.id for p in purchases])
-                    Purchase.process(purchases)
-
-
 class OpenSupplier(Wizard):
     'Open Suppliers'
     __name__ = 'purchase.open_supplier'
@@ -1817,9 +1326,12 @@ class OpenSupplier(Wizard):
         pool = Pool()
         ModelData = pool.get('ir.model.data')
         Wizard = pool.get('ir.action.wizard')
+        Purchase = pool.get('purchase.purchase')
         cursor = Transaction().cursor
+        purchase = Purchase.__table__()
 
-        cursor.execute("SELECT DISTINCT(party) FROM purchase_purchase")
+        cursor.execute(*purchase.select(purchase.party,
+                group_by=purchase.party))
         supplier_ids = [line[0] for line in cursor.fetchall()]
         action['pyson_domain'] = PYSONEncoder().encode(
             [('id', 'in', supplier_ids)])
@@ -1848,11 +1360,15 @@ class HandleShipmentExceptionAsk(ModelView):
     @classmethod
     def __register__(cls, module_name):
         cursor = Transaction().cursor
+        model = Table('ir_model')
         # Migration from 1.2: packing renamed into shipment
-        cursor.execute("UPDATE ir_model "
-            "SET model = REPLACE(model, 'packing', 'shipment') "
-            "WHERE model like '%%packing%%' AND module = %s",
-            (module_name,))
+        cursor.execute(*model.update(
+                columns=[model.model],
+                values=[Overlay(model.model, 'shipment',
+                        Position('packing', model.model),
+                        len('packing'))],
+                where=model.model.like('%packing%')
+                & (model.module == module_name)))
         super(HandleShipmentExceptionAsk, cls).__register__(module_name)
 
 
